proc warn { msg } {
  set w [genName .warn]
  toplevel $w
  frame $w.m
  pack append $w $w.m {top expand fill}

  label $w.m.l -text "Warning: $msg"
  button $w.m.b -command "destroy $w" -text "OK"

  pack append $w.m $w.m.l top $w.m.b top
}

proc genName {{var var}} {
  global nameCount
  if [info exists nameCount($var)] {
    incr nameCount($var)
  } else {
    set nameCount($var) 0
  }
  return "$var$nameCount($var)"
}

proc frameMultChoice {frame msg choices default} {
  frame $frame
  subMultChoice $frame $msg $choices $default
}

proc multChoice {title msg choices default} {
  set w [genName .mult]
  toplevel $w
  wm title $w $title
  subMultChoice $w $msg $choices $default
}

proc subMultChoice {w msg choices default} {
  set var [genName]
  frame $w.m -relief raised -border 1
  frame $w.b -relief raised -border 1
  set style {top fill expand}
  if {$msg != ""} {
    pack append $w $w.m $style
  }
  pack append $w $w.b $style
  label $w.m.l -text $msg
  pack append $w.m $w.m.l top
  foreach i $choices {
    set text [lindex $i 0]
    set command [lindex $i 1]
    set style1 {left expand padx 40 pady 20}
    if {$text == $default} {
      set style2 {expand padx 5 pady 5}
      set button default
    } {
      set style2 {expand}
      set button [genName button]
    }
    frame $w.b.$button -relief sunken -border 1
    pack append $w.b $w.b.$button $style1
    button $w.b.$button.b -text $text -command "destroy $w; $command"
    pack append $w.b.$button $w.b.$button.b $style2
    bind $w.b.$button <Enter> "$w.b.$button.b activate"
    bind $w.b.$button <Leave> "$w.b.$button.b deactivate"
    bind $w.b.$button <Return> "$w.b.$button.b invoke"
    bind $w.b.$button.b <Return> "$w.b.$button.b invoke"
  }
  bind $w.m <Enter>	"$w.b.default.b activate"
  bind $w.m.l <Enter>	"$w.b.default.b activate"
  bind $w.b <Enter>	"$w.b.default.b activate"
  bind $w.m <Leave>	"$w.b.default.b deactivate"
  bind $w.m.l <Leave>	"$w.b.default.b deactivate"
  bind $w.b <Leave>	"$w.b.default.b deactivate"
  bind $w.m <Return>	"$w.b.default.b invoke"
  bind $w.m.l <Return>	"$w.b.default.b invoke"
  bind $w.b <Return>	"$w.b.default.b invoke"
  eval [list proc $w.invoke_default {} [list $w.b.default.b invoke]]
}

proc yesNo { title msg yesAction noAction default} {
  multChoice $title $msg [list "Yes [list $yesAction]" "No [list $noAction]"] $default
}

proc quit { prog } {
  #multChoice "Quit" "Are you sure you want to quit $prog?" {{Yes "destroy ."} {No ""}} No
  yesNo "Quit" "Are you sure you want to quit $prog?" "destroy ." "noop" No
}

proc noop {} {
}  

proc toggle {var} {
  global $var
  if [info exists $var] {
    if [set $var] {
      set $var 0
    } else {
      set $var 1
    }
  } else {
    set $var 1
  }
}
