proc makeBoard {f width height} {
  global $f.white.IsRobot $f.black.IsRobot
  frame $f.menu
  frame $f.msg
  frame $f.board
  frame $f.stat -bd 10
  pack append $f $f.menu {top frame w} $f.msg top $f.stat top $f.board top

  menubutton $f.menu.mb -text "Options" -menu $f.menu.m -relief raised
  pack append $f.menu $f.menu.mb left
  menu $f.menu.m
  $f.menu.m add command -label "Hint" -command "hint $f"
  $f.menu.m add command -label "Back Up" -command "backup $f"
  $f.menu.m add command -label "Reset Board..." -command "askResetBoard $f"
  $f.menu.m add command -label "Reset Board Size..." -command "askResetBoardSize $f"
  $f.menu.m add checkbutton -label "White machine player" -var $f.white.IsRobot -command "setRobot $f white"
  $f.menu.m add checkbutton -label "Black machine player" -var $f.black.IsRobot -command "setRobot $f black"
  $f.menu.m add command -label "Quit..." -command "quit Reversi"

  label $f.msg.l1
  label $f.msg.l2
  label $f.stat.l1
  label $f.stat.l2
  label $f.stat.l3
  pack append $f.msg $f.msg.l1 left $f.msg.l2 left
  pack append $f.stat $f.stat.l1 top $f.stat.l2 top $f.stat.l3 top
  makeButtons $f $width $height
  set $f.white.IsRobot 0
  set $f.black.IsRobot 1
  resetBoard $f
}

proc makeButtons {f width height} {
  global $f.width $f.height $f.maxMoves $f.moveCount
  set $f.maxMoves [expr {$width * $height}]
  set $f.moveCount 0
  set $f.width $width
  set $f.height $height
  for {set i 0} {$i < $height} {incr i} {
    frame $f.board.$i
    pack append $f.board $f.board.$i top
    for {set j 0} {$j < $width} {incr j} {
      button $f.board.$i.$j -command "after 10 move $f $i $j"
      pack append $f.board.$i $f.board.$i.$j left
      global $f.$i.$j.white.nbr
      set $f.$i.$j.white.nbr 0
      global $f.$i.$j.black.nbr
      set $f.$i.$j.black.nbr 0
      global $f.$i.$j.value
      set $f.$i.$j.value [calcValue $f $i $j]
    }
  }
}

proc askResetBoard {f} {
  yesNo "Reset Reversi Board" "Are you sure you want to reset the board?" "resetBoard $f" "noop" No
}

proc askResetBoardSize {f} {
  global $f.width $f.height
  set w [genName .resetSize]
  toplevel $w
  wm title $w "Reset Board Size"
  frame $w.ask

  label $w.ask.hl -text "Height:"
  label $w.ask.wl -text "Width:"
  menubutton $w.ask.height -width 2 -text [set $f.height] -menu $w.ask.hmenu -relief raised
  menubutton $w.ask.width -width 2 -text [set $f.width] -menu $w.ask.wmenu -relief raised
  pack append $w.ask $w.ask.hl left $w.ask.height {left padx 10} $w.ask.wl left $w.ask.width {left padx 10}

  menu $w.ask.hmenu
  menu $w.ask.wmenu
  foreach i {6 8 10 12} {
    $w.ask.hmenu add command -label $i -command "$w.ask.height config -text $i"
    $w.ask.wmenu add command -label $i -command "$w.ask.width config -text $i"
  }

  frameMultChoice $w.yesNo "" [list [list OK "resetBoardSize $w.ask $f;destroy $w"] [list Cancel "destroy $w"]] Cancel
  pack append $w $w.ask {top fillx pady 30} $w.yesNo {top fillx}
}

proc resetBoardSize {ask f} {
  destroy $f
  toplevel $f
  wm title $f Reversi
  set width [lindex [$ask.width config -text] 4]
  set height [lindex [$ask.height config -text] 4]
  makeBoard $f $width $height
}

proc resetBoard {f} {
  global $f.width $f.height
  set width [set $f.width]
  set height [set $f.height]
  for {set i 0} {$i < $height} {incr i} {
    for {set j 0} {$j < $width} {incr j} {
      makeEmpty $f $i $j
    }
  }
  setBoard $f
}

proc flashSquare {f row col} {
  #set orig [get $f $row $col]
  #makeBlack $f $row $col; update
  #makeWhite $f $row $col; update
  #makeBlack $f $row $col; update
  #makeWhite $f $row $col; update
  #makeColor $f $row $col $orig; update
  #$f.board.$row.$col config -rel sunken
  #after 1000 $f.board.$row.$col config -rel raised
  $f.board.$row.$col flash
  $f.board.$row.$col activate
  after 4000 $f.board.$row.$col deactivate
}

proc setBoard {f} {
  global $f.moveList $f.maxMoves $f.width $f.height $f.turn $f.bestMove
  global $f.white.count $f.black.count
  set width [set $f.width]
  set height [set $f.height]
  set centerX [expr {$height / 2}]
  set centerY [expr {$width / 2}]
  set $f.white.count 0
  set $f.black.count 0
  makeBlack $f [expr {$centerX - 1}] $centerY
  makeBlack $f $centerX [expr {$centerY - 1}]
  makeWhite $f [expr {$centerX - 1}] [expr {$centerY - 1}]
  makeWhite $f $centerX $centerY
  set $f.turn white
  incr $f.maxMoves -4
  set $f.moveList {}
  set $f.bestMove [chooseMove $f [set $f.turn]]
  setStatLines $f
  setRobot $f white
}

proc setRobot {f color} {
  global $f.$color.IsRobot $f.turn $f.bestMove
  set turn [set $f.turn]
  if {! [set $f.$color.IsRobot]} {return}
  if {$turn != $color} {return}
  eval [concat move $f [set $f.bestMove]]
  update
}

proc globIncr {var {inc 1}} {
  global $var
  if [info exists $var] {incr $var $inc}
}

proc doIncr {f row col color {inc 1}} {
  globIncr $f.$color.count $inc
  globIncr $f.[expr {$row + 1}].[expr {$col - 1}].$color.nbr $inc
  globIncr $f.[expr {$row + 1}].$col.$color.nbr $inc
  globIncr $f.[expr {$row + 1}].[expr {$col + 1}].$color.nbr $inc
  globIncr $f.$row.[expr {$col - 1}].$color.nbr $inc
  globIncr $f.$row.[expr {$col + 1}].$color.nbr $inc
  globIncr $f.[expr {$row - 1}].[expr {$col - 1}].$color.nbr $inc
  globIncr $f.[expr {$row - 1}].$col.$color.nbr $inc
  globIncr $f.[expr {$row - 1}].[expr {$col + 1}].$color.nbr $inc
}

proc makeWhite {f row col} {
  global sourceDir
  global $f.white.count $f.black.count
  global $f.board.$row.$col.color
  if {[get $f $row $col] == "black"} {doIncr $f $row $col black -1; doIncr $f $row $col white}
  if {[get $f $row $col] == "empty"} {doIncr $f $row $col white}
  set $f.board.$row.$col.color white
  $f.board.$row.$col config -bitmap "@$sourceDir/bitmaps/white.bm"
}

proc makeBlack {f row col} {
  global sourceDir
  global $f.white.count $f.black.count
  global $f.board.$row.$col.color
  if {[get $f $row $col] == "white"} {doIncr $f $row $col white -1; doIncr $f $row $col black}
  if {[get $f $row $col] == "empty"} {doIncr $f $row $col black}
  set $f.board.$row.$col.color black
  $f.board.$row.$col config -bitmap "@$sourceDir/bitmaps/black.bm"
}

proc makeEmpty {f row col} {
  global sourceDir
  global $f.white.count $f.black.count
  global $f.board.$row.$col.color
  if {[get $f $row $col] == "black"} {incr $f.black.count -1}
  if {[get $f $row $col] == "white"} {incr $f.white.count -1}
  set $f.board.$row.$col.color empty
  $f.board.$row.$col config -bitmap "@$sourceDir/bitmaps/empty.bm"
}

proc makeColor {f row col color} {
  if {$color == "black"} {makeBlack $f $row $col}
  if {$color == "white"} {makeWhite $f $row $col}
  if {$color == "empty"} {makeEmpty $f $row $col}
}

proc get {f row col} {
  global $f.board.$row.$col.color
  if {! [info exists $f.board.$row.$col.color]} {return "empty"}
  return [set $f.board.$row.$col.color]
}

proc getValue {f row col} {
  global $f.$row.$col.value
  #if {! [info exists $f.$row.$col.value]} {return 0}
  return [set $f.$row.$col.value]
}

proc calcValue {f row col} {
  global $f.width $f.height
  set value 0
  set width [set $f.width]
  set height [set $f.height]
  set h [min $width $height]
  set distToHEdge [min [expr {$height - $row - 1}] $row]
  set distToVEdge [min [expr {$width - $col - 1}] $col]
  set distToEdge [min $distToHEdge $distToVEdge]
  set distToCorner [max $distToHEdge $distToVEdge]

  set oddDistToCorner [expr {$distToCorner % 2}]

  if {$oddDistToCorner == 1} {
    incr value -2
  } else {
    incr value 2
  }
  if {$distToEdge == 0} {
    incr value $h
  } else {
    if {$distToEdge == 1} {
      incr value -$h
    } else {
      incr value -2
    }
  }
  if {$distToCorner == 0} {
    incr value $h
  }
  return $value
}

proc flip {f row col} {
  set val [get $f $row $col]
  if {$val == "black"} {
    makeWhite $f $row $col
    update
    makeBlack $f $row $col
    update
    makeWhite $f $row $col
  } else {
    makeBlack $f $row $col
    update
    makeWhite $f $row $col
    update
    makeBlack $f $row $col
  }
  # Make the flipping look pretty
  update
}

proc flipList {f flips} {
  foreach spot $flips {
    flip $f [lindex $spot 0] [lindex $spot 1]
  }
}

proc opposite {color} {
  if {$color == "white"} {return "black"}
  if {$color == "black"} {return "white"}
  if {$color == "empty"} {return "empty"}
}

proc whichFlip {f row col color} {
  global $f.width $f.height
  global $f.$row.$col.[opposite $color].nbr
  set width [set $f.width]
  set height [set $f.height]
  if {[set $f.$row.$col.[opposite $color].nbr] == 0} {return {}}
  if {[get $f $row $col] != "empty"} {return {}}
  set flip {}
  # up
  set couldFlip {}
  for {set i [expr {$row - 1}]} {$i >= 0} {incr i -1} {
    if {$color == [opposite [get $f $i $col]]} {
      lappend couldFlip [list $i $col]
    } else { if {$color == [get $f $i $col]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $i $col] == "empty"} {
      break
    }}}
  }
  # down
  set couldFlip {}
  for {set i [expr {$row + 1}]} {$i < $height} {incr i} {
    if {$color == [opposite [get $f $i $col]]} {
      lappend couldFlip [list $i $col]
    } else { if {$color == [get $f $i $col]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $i $col] == "empty"} {
      break
    }}}
  }
  # left
  set couldFlip {}
  for {set i [expr {$col - 1}]} {$i >= 0} {incr i -1} {
    if {$color == [opposite [get $f $row $i]]} {
      lappend couldFlip [list $row $i]
    } else { if {$color == [get $f $row $i]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $row $i] == "empty"} {
      break
    }}}
  }
  # right
  set couldFlip {}
  for {set i [expr {$col + 1}]} {$i < $width} {incr i} {
    if {$color == [opposite [get $f $row $i]]} {
      lappend couldFlip [list $row $i]
    } else { if {$color == [get $f $row $i]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $row $i] == "empty"} {
      break
    }}}
  }
  # upleft
  set couldFlip {}
  for {set i [expr {$row - 1}]; set j [expr {$col - 1}]} {$i >= 0 && $j >= 0} {incr i -1; incr j -1} {
    if {$color == [opposite [get $f $i $j]]} {
      lappend couldFlip [list $i $j]
    } else { if {$color == [get $f $i $j]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $i $j] == "empty"} {
      break
    }}}
  }
  # downleft
  set couldFlip {}
  for {set i [expr {$row + 1}]; set j [expr {$col - 1}]} {$i < $height && $j >= 0} {incr i; incr j -1} {
    if {$color == [opposite [get $f $i $j]]} {
      lappend couldFlip [list $i $j]
    } else { if {$color == [get $f $i $j]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $i $j] == "empty"} {
      break
    }}}
  }
  # upright
  set couldFlip {}
  for {set i [expr {$row - 1}]; set j [expr {$col + 1}]} {$i >= 0 && $j < $width} {incr i -1; incr j} {
    if {$color == [opposite [get $f $i $j]]} {
      lappend couldFlip [list $i $j]
    } else { if {$color == [get $f $i $j]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $i $j] == "empty"} {
      break
    }}}
  }
  # downright
  set couldFlip {}
  for {set i [expr {$row + 1}]; set j [expr {$col + 1}]} {$i < $height && $j < $width} {incr i; incr j} {
    if {$color == [opposite [get $f $i $j]]} {
      lappend couldFlip [list $i $j]
    } else { if {$color == [get $f $i $j]} {
      set flip [concat $flip $couldFlip]
      break
    } else { if {[get $f $i $j] == "empty"} {
      break
    }}}
  }
  return $flip
}

proc hint {f} {
  global $f.bestMove
  set move [set $f.bestMove]
  if {$move == {}} {
    $f.stat.l1 config -text "No moves available!"
    return
  }
  flashSquare $f [lindex $move 0] [lindex $move 1]
}

proc distToEdge {f row col} {
  global $f.width $f.height
  set width [set $f.width]
  set height [set $f.height]
  return [min [expr {$height - $row - 1}] [expr {$width - $col - 1}] $row $col]
}

proc chooseMove {f turn} {
  global $f.width $f.height
  set width [set $f.width]
  set height [set $f.height]
  set bestMove {}
  set bestValue -999
  for {set i 0} {$i < $height} {incr i} {
    for {set j 0} {$j < $width} {incr j} {
      set flips [whichFlip $f $i $j $turn]
      if {[llength $flips] == 0} {continue}
      #set value [expr {[llength $flips] + [getValue $f $i $j]}]
      set value [llength $flips]
      set dist [distToEdge $f $i $j]
      if {$dist == 0} {
        set value [expr {$value + [max $height $width]}]
      } else { if {$dist == 1} {
        set value [expr {$value - 1}]
      } else {
        set value [expr {$value + $dist}]
      }}
      if {$value > $bestValue || ([random 3] == 0 && $value == $bestValue)} {
          set bestValue $value
          set bestMove "$i $j"
      }
    }
  }
  return $bestMove
}

proc backup {f {count 1}} {
  global $f.moveList $f.moveCount $f.turn $f.bestMove
  set moveList [set $f.moveList]
  set moveCount [set $f.moveCount]
  if {$moveCount == 0} {
    $f.stat.l1 config -text "Can't back up any further!"
    return
  }
  set lastTurn [lindex $moveList [expr {$moveCount * 3 - 1}]]
  set lastFlipList [lindex $moveList [expr {$moveCount * 3 - 2}]]
  set lastMove [lindex $moveList [expr {$moveCount * 3 - 3}]]
  set moveList [lrange $moveList 0 [expr {$moveCount * 3 - 4}]]
  makeEmpty $f [lindex $lastMove 0] [lindex $lastMove 1]
  flipList $f $lastFlipList
  set $f.moveList $moveList
  incr $f.moveCount -1
  set $f.turn $lastTurn
  set $f.bestMove [chooseMove $f [set $f.turn]]
  setStatLines $f
}

proc postWinner {f} {
  global $f.white.count $f.black.count $f.bestMove
  set $f.bestMove {}
  if {[set $f.white.count] > [set $f.black.count]} {
    set msg "White wins, [set $f.white.count] to [set $f.black.count]!!!"
  }
  if {[set $f.white.count] < [set $f.black.count]} {
    set msg "Black wins, [set $f.black.count] to [set $f.white.count]!!!"
  }
  if {[set $f.white.count] == [set $f.black.count]} {
    set msg "Ugh. It's a tie, [set $f.white.count] all."
  }
  multChoice "Game Over!" $msg [list {OK noop} [list "Reset Board" "resetBoard $f"] [list "Quit" "destroy [winfo parent $f]"]] "Reset Board"
  update
}

proc move {f row col} {
  global $f.turn $f.white.IsRobot $f.black.IsRobot $f.moveCount $f.maxMoves $f.moveList $f.bestMove
  set maxMoves [set $f.maxMoves]
  set turn [set $f.turn]
  set flips [whichFlip $f $row $col $turn]
  if {[llength $flips] > 0} {
    if {$turn == "white"} {
      makeWhite $f $row $col
    } else {
      makeBlack $f $row $col
    }
    flashSquare $f $row $col
    flipList $f $flips
    lappend $f.moveList [list $row $col] $flips $turn
    #echo [set $f.moveList]
    incr $f.moveCount
    set turn [opposite $turn]
    set $f.turn $turn
    $f.stat.l1 config -text ""
  } else {
    $f.stat.l1 config -text "Illegal move!"
    return
  }
  setStatLines $f
  if {[set $f.moveCount] == $maxMoves} {
    $f.stat.l1 config -text "Game over, man!"
    postWinner $f
    return
  }
  if {[set $f.$turn.IsRobot]} {
    $f.stat.l1 config -text "$turn is thinking..."
  }
  update
  after 10 nextMove $f
}

proc nextMove {f} {
  global $f.bestMove $f.turn $f.white.IsRobot $f.black.IsRobot
  set turn [set $f.turn]
  set $f.bestMove [chooseMove $f [set $f.turn]]
  if {[set $f.bestMove] == {}} {
    $f.stat.l1 config -text "$turn can't move!"
    set turn [opposite $turn]
    set $f.turn $turn
    set $f.bestMove [chooseMove $f $turn]
    setStatLines $f
    if {[set $f.bestMove] == {}} {
      $f.stat.l1 config -text "No moves possible!"
      postWinner $f
      return
    }
  }
  if {[set $f.$turn.IsRobot]} {
    update
    after 10 eval [concat move $f [set $f.bestMove]]
  }
}

proc setStatLines {f} {
  global sourceDir
  global $f.turn $f.white.count $f.black.count
  set turn [set $f.turn]
  $f.stat.l2 config -text "White: [set $f.white.count]"
  $f.stat.l3 config -text "Black: [set $f.black.count]"
  $f.msg.l2 config -bitmap "@$sourceDir/bitmaps/$turn.bm"
  $f.msg.l1 config -text "$turn's turn"
  update
}

wm withdraw .
toplevel .w
wm title .w Reversi
makeBoard .w $width $height
